from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from blackhole_io.store.models import FileRecord, FileRecordInput


class AbstractStore(ABC):
    @abstractmethod
    async def save(self, record: "FileRecordInput") -> "FileRecord":
        pass

    @abstractmethod
    async def get(self, filename: str) -> "FileRecord | None":
        pass

    @abstractmethod
    async def delete(self, filename: str) -> None:
        pass

    @abstractmethod
    async def create_tables(self) -> None:
        pass



import asyncio
import concurrent.futures
import functools
import logging
from typing import Any

import boto3
from botocore.config import Config
from starlette.datastructures import UploadFile

from blackhole_io.adapters.abstract import AbstractAdapter
from blackhole_io.blackhole_file import BlackholeFile
from blackhole_io.configs.s3 import S3Config
from blackhole_io.types import UploadFileType

logger = logging.getLogger(__name__)


class S3Adapter(AbstractAdapter):
    def __init__(self, config: S3Config, **kwargs) -> None:
        super().__init__(config)
        self.client = boto3.client(
            "s3",
            config=Config(
                signature_version="v4",
                region_name=self.config.region,
            ),
            aws_access_key_id=self.config.access_key,
            aws_secret_access_key=self.config.secret_key,
        )

    async def put(self, file: BlackholeFile) -> str:
        key = file.filename
        extra = dict(file.extra)
        logger.info("Uploading %s to bucket", key)
        loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool, functools.partial(self._sync_put, file.data_to_upload, key, **extra)
            )

    async def put_all(self, files: list[BlackholeFile]) -> list[Any]:
        return await asyncio.gather(
            *[self.put(file=file) for file in files],
        )

    async def get(self, file_name: str) -> BlackholeFile:
        loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool, functools.partial(self._sync_get, file_name)
            )

    async def exists(self, file_name: str) -> bool:
        loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool, functools.partial(self._sync_exists, file_name)
            )

    async def delete(self, file_name: str) -> None:
        loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            await loop.run_in_executor(
                pool, functools.partial(self._sync_delete, file_name)
            )

    def _sync_get(self, file_name: str) -> BlackholeFile:
        response = self.client.get_object(
            Bucket=self.config.bucket,
            Key=file_name,
        )
        data = response["Body"].read()
        content_type = response.get("ContentType", "application/octet-stream")
        size = response.get("ContentLength", len(data))

        return BlackholeFile(
            filename=file_name,
            content_type=content_type,
            size=size,
            data=data,
        )

    def _sync_exists(self, file_name: str) -> bool:
        try:
            self.client.head_object(
                Bucket=self.config.bucket,
                Key=file_name,
            )
            return True
        except Exception:
            return False

    def _sync_delete(self, file_name: str) -> None:
        self.client.delete_object(
            Bucket=self.config.bucket,
            Key=file_name,
        )

    def _sync_put(self, file: UploadFileType, key: str, **kwargs) -> str:
        bucket = kwargs.pop("Bucket", self.config.bucket)
        kwargs.pop("Key", None)

        if isinstance(file, str):
            self.client.upload_file(Filename=file, Bucket=bucket, Key=key, **kwargs)
            return key

        fileobj = file
        if isinstance(file, UploadFile):
            fileobj = file.file

        self.client.upload_fileobj(Fileobj=fileobj, Bucket=bucket, Key=key, **kwargs)
        logger.info("Uploaded %s to bucket", key)
        return key



from typing import Union

from blackhole_io.configs.gcp import GCPConfig
from blackhole_io.configs.local import LocalConfig
from blackhole_io.configs.s3 import S3Config

ConfigType = Union[S3Config, GCPConfig, LocalConfig]

__all__ = ["S3Config", "GCPConfig", "LocalConfig", "ConfigType"]



import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import yaml

from blackhole_io.configs import ConfigType
from blackhole_io.configs.gcp import GCPConfig
from blackhole_io.configs.local import LocalConfig
from blackhole_io.configs.s3 import S3Config

if TYPE_CHECKING:
    from blackhole_io.store.abstract import AbstractStore

logger = logging.getLogger(__name__)

ADAPTER_MAP: dict[str, type] = {
    "s3": S3Config,
    "gcp": GCPConfig,
    "local": LocalConfig,
}


def _build_store_from_dict(store_data: dict) -> "AbstractStore":
    from blackhole_io.store.factory import create_store
    return create_store(store_data)


def _load_from_yaml(path: Path) -> tuple[ConfigType, Optional["AbstractStore"]]:
    if not path.exists():
        raise FileNotFoundError(f"Config file not found: {path}")

    data = yaml.safe_load(path.read_text())

    store_data = data.pop("store", None)

    adapter = data.get("adapter")
    if adapter is None:
        raise ValueError("YAML config must contain 'adapter' key (s3, gcp, local)")

    config_cls = ADAPTER_MAP.get(adapter)
    if config_cls is None:
        raise ValueError(
            f"Unknown adapter '{adapter}'. Must be one of: {', '.join(ADAPTER_MAP)}"
        )

    env_prefix = config_cls.model_config.get("env_prefix", "")
    for key in config_cls.env_fields():
        if key not in data:
            env_val = os.environ.get(f"{env_prefix}{key.upper()}")
            if env_val is not None:
                data[key] = env_val

    logger.debug("Loaded '%s' config from YAML: %s", adapter, path)
    file_config = config_cls(**data)
    store = _build_store_from_dict(store_data) if store_data else None
    return file_config, store


def _discover_yaml() -> Optional[Path]:
    path = Path.cwd() / "config" / "blackhole.yaml"
    return path if path.exists() else None


SUPPORTED_EXTENSIONS = {".yaml", ".yml"}


def _load_from_file(path: Path) -> tuple[ConfigType, Optional["AbstractStore"]]:
    if path.suffix not in SUPPORTED_EXTENSIONS:
        raise ValueError(
            f"Unsupported config file format: '{path.suffix}'. "
            f"Supported: {', '.join(sorted(SUPPORTED_EXTENSIONS))}"
        )
    return _load_from_yaml(path)


def load_config(config: Optional[ConfigType | str | Path] = None) -> ConfigType:
    """Return only the file-adapter config. Use load_config_with_store for store too."""
    file_config, _ = _load_config_pair(config)
    return file_config


def load_config_with_store(
    config: Optional[ConfigType | str | Path] = None,
) -> tuple[ConfigType, Optional["AbstractStore"]]:
    """Return (file_adapter_config, store) parsed from config source."""
    return _load_config_pair(config)


def _load_config_pair(
    config: Optional[ConfigType | str | Path],
) -> tuple[ConfigType, Optional["AbstractStore"]]:
    if isinstance(config, ConfigType):
        return config, None

    if isinstance(config, (str, Path)):
        return _load_from_file(Path(config))

    # config is None — try auto-discovery
    yaml_path = _discover_yaml()
    if yaml_path is not None:
        logger.debug("Auto-discovered config at %s", yaml_path)
        return _load_from_yaml(yaml_path)

    # Fall back to env vars
    adapter = os.environ.get("BLACKHOLE_ADAPTER")
    if adapter is not None:
        logger.debug("Using env var fallback: BLACKHOLE_ADAPTER=%s", adapter)
        config_cls = ADAPTER_MAP.get(adapter)
        if config_cls is None:
            raise ValueError(
                f"Unknown adapter '{adapter}'. Must be one of: {', '.join(ADAPTER_MAP)}"
            )
        return config_cls(), None

    raise ValueError(
        "No configuration found. Provide a config object, a YAML file path, "
        "place config/blackhole.yaml in the working directory, "
        "or set BLACKHOLE_ADAPTER environment variable."
    )



from typing import Optional

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from blackhole_io.configs.abstract import AbstractConfig


class GCPConfig(AbstractConfig):
    model_config = SettingsConfigDict(env_prefix="BLACKHOLE_GCP_", extra="ignore")

    bucket: str = Field(..., description="The name of the GCP bucket")
    project_id: str = Field(..., description="The GCP project ID")
    credentials: Optional[dict] = Field(default=None, description="GCP credentials for authentication")

    @classmethod
    def env_fields(cls) -> set[str]:
        return {"credentials"}



import asyncio
import logging
from pathlib import Path
from typing import Any, Optional
from uuid import uuid4

from blackhole_io.adapters import UploadFileType
from blackhole_io.adapters.factory import AdapterFactory
from blackhole_io.blackhole_file import BlackholeFile
from blackhole_io.configs import ConfigType
from blackhole_io.configs.loader import load_config_with_store
from blackhole_io.store.abstract import AbstractStore
from blackhole_io.store.models import FileRecordInput

logger = logging.getLogger("blackhole_io")
logger.setLevel(logging.ERROR)


class Blackhole:
    def __init__(
        self,
        config: Optional[ConfigType | str | Path] = None,
        log_level: Optional[int] = None,
        store: Optional[AbstractStore] = None,
    ) -> None:
        if log_level is not None:
            logger.setLevel(log_level)
            if not logger.handlers:
                logger.addHandler(logging.StreamHandler())

        file_config, yaml_store = load_config_with_store(config)
        self.config = file_config
        self.adapter = AdapterFactory.create(self.config)
        self.store = store if store is not None else yaml_store

    # TODO: return BlackholeFile
    async def put(
        self,
        file: UploadFileType,
        filename: Optional[str] = None,
        extra_metadata: Optional[dict[str, Any]] = None,
    ) -> str:
        upload_filename = filename or uuid4().hex
        content_type = None

        if hasattr(file, "filename") and not filename:
            upload_filename = file.filename or upload_filename
            content_type = file.content_type

        bh_file = BlackholeFile(
            filename=upload_filename,
            content_type=content_type,
            data_to_upload=file,
        )

        filename = await self.adapter.put(bh_file)

        logger.info(
            f"Uploading filename: {upload_filename} with content_type: {content_type}"
        )

        if self.store is not None:
            await self.store.save(
                FileRecordInput(filename=filename, content_type=content_type, extra_metadata=extra_metadata)
            )
        return filename

    async def put_all(
        self,
        files: list[UploadFileType],
        extra_metadata: Optional[list[Optional[dict[str, Any]]]] = None,
    ) -> list[str]:
        files_to_upload = [
            BlackholeFile(filename=uuid4().hex, data_to_upload=f)
            for f in files
        ]
        filenames = await self.adapter.put_all(files_to_upload)
        if self.store is not None:
            meta_list = extra_metadata or [None] * len(filenames)
            await asyncio.gather(
                *[
                    self.store.save(
                        FileRecordInput(filename=fn, extra_metadata=m)
                    )
                    for fn, m in zip(filenames, meta_list)
                ]
            )
        return filenames

    async def get(self, file_name: str) -> BlackholeFile:
        return await self.adapter.get(file_name)

    async def exists(self, file_name: str) -> bool:
        return await self.adapter.exists(file_name)

    async def delete(self, file_name: str) -> None:
        await self.adapter.delete(file_name)



from typing import Optional

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from blackhole_io.configs.abstract import AbstractConfig


class S3Config(AbstractConfig):
    model_config = SettingsConfigDict(env_prefix="BLACKHOLE_S3_", extra="ignore")

    bucket: str = Field(..., description="The name of the S3 bucket")
    region: str = Field(
        ..., description="The AWS region where the S3 bucket is located"
    )
    access_key: Optional[str] = Field(default=None, description="AWS access key for authentication")
    secret_key: Optional[str] = Field(default=None, description="AWS secret key for authentication")

    @classmethod
    def env_fields(cls) -> set[str]:
        return {"access_key", "secret_key"}



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from blackhole_io.adapters.abstract import AbstractAdapter
from blackhole_io.configs import GCPConfig, LocalConfig, S3Config

if TYPE_CHECKING:
    from blackhole_io.adapters.gcp_adapter import GCPAdapter
    from blackhole_io.adapters.local_adapter import LocalAdapter
    from blackhole_io.adapters.s3_adapter import S3Adapter
    from blackhole_io.configs import ConfigType


class AdapterFactory:
    @overload
    @classmethod
    def create(cls, config: S3Config) -> "S3Adapter": ...

    @overload
    @classmethod
    def create(cls, config: GCPConfig) -> "GCPAdapter": ...

    @overload
    @classmethod
    def create(cls, config: LocalConfig) -> "LocalAdapter": ...

    @classmethod
    def create(cls, config: ConfigType) -> AbstractAdapter:
        if isinstance(config, S3Config):
            from blackhole_io.adapters.s3_adapter import S3Adapter
            return S3Adapter(config)

        if isinstance(config, GCPConfig):
            from blackhole_io.adapters.gcp_adapter import GCPAdapter
            return GCPAdapter(config)

        if isinstance(config, LocalConfig):
            from blackhole_io.adapters.local_adapter import LocalAdapter
            return LocalAdapter(config)

        raise ValueError("Unsupported configuration type")



import asyncio
from io import BytesIO
from pathlib import Path

import aiofiles
import aiofiles.os
from starlette.datastructures import UploadFile

from blackhole_io.adapters.abstract import AbstractAdapter
from blackhole_io.blackhole_file import BlackholeFile

# TODO: add support for directory structure - now put supports only flat structure

class LocalAdapter(AbstractAdapter):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    async def put(self, file: BlackholeFile) -> str:
        directory = Path(self.config.directory)
        file_id = file.filename
        data_to_upload = file.data_to_upload

        if isinstance(data_to_upload, str):
            filename = directory / (file_id + Path(data_to_upload).suffix)
        else:
            # for UploadFile we probably have a correct content type even if its no extension at all
            # bytes and BytesIO just rely on file_id to determine the extension
            filename = directory / file_id

        if isinstance(data_to_upload, str):
            async with aiofiles.open(data_to_upload, "rb") as f:
                data = await f.read()
            async with aiofiles.open(filename, "wb") as out:
                await out.write(data)
        elif isinstance(data_to_upload, bytes):
            async with aiofiles.open(filename, "wb") as out:
                await out.write(data_to_upload)
        elif isinstance(data_to_upload, BytesIO):
            async with aiofiles.open(filename, "wb") as out:
                await out.write(data_to_upload.getvalue())
        elif isinstance(data_to_upload, UploadFile):
            filebytes = await data_to_upload.read()
            async with aiofiles.open(filename, "wb") as out:
                await out.write(filebytes)
        else:
            raise TypeError("Unsupported file type. Must be str, bytes, BytesIO or UploadFile.")

        return filename.name

    async def put_all(self, files: list[BlackholeFile]) -> list[str]:
        return await asyncio.gather(*[self.put(file) for file in files])

    async def get(self, file_name: str) -> BlackholeFile:
        file_path = Path(self.config.directory) / file_name

        if not await aiofiles.os.path.exists(file_path):
            raise FileNotFoundError(f"File {file_path} does not exist.")
        if not await aiofiles.os.path.isfile(file_path):
            raise IsADirectoryError(f"{file_path} is a directory, not a file.")
        if not file_path.stat().st_mode & 0o444:
            raise PermissionError(f"File {file_path} is not readable.")

        async with aiofiles.open(file_path, "rb") as f:
            data = await f.read()

        return BlackholeFile(
            filename=file_name,
            size=len(data),
            data=data,
        )

    async def exists(self, file_name: str) -> bool:
        file_path = Path(self.config.directory) / file_name
        return await aiofiles.os.path.isfile(file_path)

    async def delete(self, file_name: str) -> None:
        file_path = Path(self.config.directory) / file_name

        if not await aiofiles.os.path.exists(file_path):
            raise FileNotFoundError(f"File {file_path} does not exist.")
        if not await aiofiles.os.path.isfile(file_path):
            raise IsADirectoryError(f"{file_path} is a directory, not a file.")
        if not file_path.stat().st_mode & 0o222:
            raise PermissionError(f"File {file_path} is not writable.")

        await aiofiles.os.remove(file_path)



from typing import Optional

from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlmodel import SQLModel, select
from sqlmodel.ext.asyncio.session import AsyncSession

from blackhole_io.store.abstract import AbstractStore
from blackhole_io.store.models import FileRecord, FileRecordInput


class SQLStore(AbstractStore):
    def __init__(
        self,
        engine: Optional[AsyncEngine] = None,
        dsn: Optional[str] = None,
    ) -> None:
        if engine is None and dsn is None:
            raise ValueError("SQLStore requires either 'engine' or 'dsn'")
        self._engine = engine or create_async_engine(dsn, echo=False)

    @property
    def engine(self) -> AsyncEngine:
        return self._engine

    async def save(self, record: FileRecordInput) -> FileRecord:
        async with AsyncSession(self._engine) as session:
            db_record = FileRecord(
                filename=record.filename,
                content_type=record.content_type,
                size=record.size,
                extra_metadata=record.extra_metadata,
            )
            session.add(db_record)
            await session.commit()
            await session.refresh(db_record)
            return db_record

    async def get(self, filename: str) -> Optional[FileRecord]:
        async with AsyncSession(self._engine) as session:
            result = await session.exec(
                select(FileRecord).where(FileRecord.filename == filename)
            )
            return result.first()

    async def delete(self, filename: str) -> None:
        async with AsyncSession(self._engine) as session:
            result = await session.exec(
                select(FileRecord).where(FileRecord.filename == filename)
            )
            record = result.first()
            if record is not None:
                await session.delete(record)
                await session.commit()

    async def create_tables(self) -> None:
        async with self._engine.begin() as conn:
            await conn.run_sync(SQLModel.metadata.create_all)



from pydantic import Field
from pydantic_settings import SettingsConfigDict

from blackhole_io.configs.abstract import AbstractConfig


class LocalConfig(AbstractConfig):
    model_config = SettingsConfigDict(env_prefix="BLACKHOLE_LOCAL_", extra="ignore")

    directory: str = Field(..., description="The name of the local directory")

    @classmethod
    def env_fields(cls) -> set[str]:
        return set()



### BlackholeFile
# Object of this class represents the information about uploaded/downloaded file

from pydantic import BaseModel, ConfigDict, Field
from typing import Optional

from blackhole_io.types import UploadFileType


class BlackholeFile(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    filename: str = Field(..., description="The name/key of the file in storage")
    content_type: Optional[str] = Field(None, description="The MIME type of the file")
    size: int = Field(default=0)
    data: bytes = Field(default=b"", repr=False)
    data_to_upload: UploadFileType = Field(default=b"", repr=False)
    extra: dict = Field(default_factory=dict)

    @property
    def blob(self) -> bytes:
        return self.data



from abc import ABC, abstractmethod
from typing import Any

from blackhole_io.blackhole_file import BlackholeFile


class AbstractAdapter(ABC):
    def __init__(self, config: Any) -> None:
        self.config = config

    @abstractmethod
    async def put(self, file: BlackholeFile) -> str:
        pass

    @abstractmethod
    async def put_all(self, files: list[BlackholeFile]) -> list[str]:
        pass

    @abstractmethod
    async def get(self, **kwargs) -> BlackholeFile:
        pass

    @abstractmethod
    async def exists(self, **kwargs) -> bool:
        pass

    @abstractmethod
    async def delete(self, **kwargs) -> None:
        pass



from typing import Any

from blackhole_io.store.abstract import AbstractStore


def create_store(store_config: dict[str, Any]) -> AbstractStore:
    store_type = store_config.get("type")
    if store_type == "sql":
        from blackhole_io.store.sql_store import SQLStore

        engine = store_config.get("engine")
        dsn = store_config.get("dsn")
        return SQLStore(engine=engine, dsn=dsn)

    raise ValueError(
        f"Unknown store type '{store_type}'. Supported types: sql"
    )



from pydantic_settings import BaseSettings, SettingsConfigDict


class AbstractConfig(BaseSettings):
    model_config = SettingsConfigDict(extra="ignore")

    @classmethod
    def env_fields(cls) -> set[str]:
        """Return field names that should be loaded from env vars, not YAML."""
        return set()



import logging

from blackhole_io.adapters.abstract import AbstractAdapter
from blackhole_io.blackhole_file import BlackholeFile

logger = logging.getLogger(__name__)


class GCPAdapter(AbstractAdapter):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    async def put(self, file: BlackholeFile) -> str:
        logger.info("PUT")
        return ""

    async def put_all(self, files: list[BlackholeFile]) -> list[str]:
        logger.info("PUT ALL")
        return [""] * len(files)

    async def get(self, file_name: str) -> BlackholeFile:
        logger.info("GET")
        return BlackholeFile(filename=file_name)

    async def exists(self, file_name: str) -> bool:
        logger.info("EXISTS")
        return False

    async def delete(self, file_name: str) -> None:
        logger.info("DELETE")



import asyncio
from typing import Optional

import typer

app = typer.Typer(name="blackhole", help="Blackhole CLI")


def _run(coro):
    asyncio.run(coro)


@app.command("create-tables")
def create_tables(
    dsn: Optional[str] = typer.Option(None, "--dsn", help="Database DSN string"),
    config: Optional[str] = typer.Option(
        None, "--config", "-c", help="Path to blackhole YAML config file"
    ),
):
    """Create database tables for the configured store."""
    from blackhole_io.store.sql_store import SQLStore

    store: Optional[object] = None

    if config is not None:
        from blackhole_io.configs.loader import load_config_with_store

        _, store = load_config_with_store(config)
        if store is None:
            typer.echo(
                "No store section found in config. Add a 'store' key to your YAML.",
                err=True,
            )
            raise typer.Exit(code=1)
    elif dsn is not None:
        store = SQLStore(dsn=dsn)
    else:
        typer.echo("Provide --dsn or --config.", err=True)
        raise typer.Exit(code=1)

    async def _create():
        await store.create_tables()
        typer.echo("Tables created successfully.")

    _run(_create())



from datetime import datetime, timezone
from typing import Any, Optional

from pydantic import BaseModel
from sqlmodel import JSON, Column, Field, SQLModel


class FileRecord(SQLModel, table=True):
    __tablename__ = "blackhole_files"

    id: Optional[int] = Field(default=None, primary_key=True)
    filename: str = Field(index=True, unique=True)
    content_type: Optional[str] = Field(default=None)
    size: int = Field(default=0)
    created_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    extra_metadata: Optional[dict[str, Any]] = Field(
        default=None, sa_column=Column("metadata", JSON)
    )


class FileRecordInput(BaseModel):
    filename: str
    content_type: str = "application/octet-stream"
    size: int = 0
    extra_metadata: Optional[dict[str, Any]] = None



from blackhole_io.store.abstract import AbstractStore
from blackhole_io.store.factory import create_store
from blackhole_io.store.models import FileRecord, FileRecordInput

__all__ = ["AbstractStore", "FileRecord", "FileRecordInput", "create_store"]
